"""app objects for vcsfile web interface"""

from cubicweb.web import uicfg
from cubicweb.web.formfields import EditableFileField
from cubicweb.web import formwidgets as wdgs

# hidden fields in edition
_afs = uicfg.autoform_section
_afs.tag_attribute(('Revision', 'revision'), 'generated')
_afs.tag_attribute(('Revision', 'changeset'), 'generated')
_afs.tag_subject_of(('*', 'content_for', 'VersionedFile'), 'generated')
_afs.tag_object_of(('*', 'content_for', 'VersionedFile'), 'generated')

# widgets / fields adjustement
_aff = uicfg.autoform_field
_affk = uicfg.autoform_field_kwargs
_aff.tag_attribute(('VersionContent', 'data'), EditableFileField)
_affk.tag_attribute(('VersionContent', 'data'), {'required': True})
_affk.tag_attribute(('Repository', 'path'), {'widget': wdgs.TextInput})
_affk.tag_attribute(('Repository', 'subpath'), {'widget': wdgs.TextInput})
_affk.tag_attribute(('VersionedFile', 'directory'), {'widget': wdgs.TextInput})
_affk.tag_attribute(('VersionedFile', 'name'), {'widget': wdgs.TextInput})

# inlined relation in edition
_afii = uicfg.autoform_is_inlined
_afii.tag_object_of(('VersionContent', 'from_revision', '*'), True)

# primary view tweaks
_pvs = uicfg.primaryview_section
_pvdc = uicfg.primaryview_display_ctrl
# internal purpose relation
_pvs.tag_subject_of(('*', 'at_revision', '*'), 'hidden')
_pvs.tag_object_of(('*', 'at_revision', '*'), 'hidden')
# displayed in attributes/relations section of Revision primary view
_pvs.tag_subject_of(('*', 'parent_revision', '*'), 'hidden')
_pvs.tag_object_of(('*', 'parent_revision', '*'), 'hidden')
_pvs.tag_object_of(('*', 'from_revision', '*'), 'hidden')
_pvdc.tag_subject_of(('*', 'from_revision', '*'), {'vid': 'incontext'})
_pvdc.tag_subject_of(('*', 'from_repository', '*'), {'vid': 'incontext'})
_pvdc.tag_subject_of(('*', 'content_for', '*'), {'vid': 'incontext'})

# we don't want automatic addrelated action for the following relations...
_abaa = uicfg.actionbox_appearsin_addmenu
for rtype in ('from_repository', 'from_revision', 'content_for',
              'parent_revision'):
    _abaa.tag_object_of(('*', rtype, '*'), False)



import os
from os.path import join, dirname
from cubicweb.server.sqlutils import sqlexec
from cubes.vcsfile.vcssource import get_vcs_source

vcssource = get_vcs_source(repo)

# remove NOT NULL on VersionContent.data 
sync_schema_props_perms(('VersionContent', 'data', 'Bytes'), syncperms=False)

# dump the sqlite database
if confirm('backport sqlite database content?'):
    dumpfile = '/tmp/vcsfile.%s.sql' % config.appid
    dumpfile2 = '/tmp/vcsfile.%s.data.sql' % config.appid
    if os.system("sqlite3 %s .dump > %s" % (vcssource.dbpath, dumpfile)):
        import sys
        print 'error while dumping sqlite database'
        print 'is sqlite3 installed?'
        sys.exit(1)

    # filter out schema statement, import only 'data' statements (eg INSERT) into
    # the system database
    session.set_pool()
    cu = session.system_sql
    status = None # None | 'data' | 'schema'
    columns = []
    stmt = None

    def process_line(line, nextline):
        global cu, status, columns, stmt
        if status is None:
            stmttype = statement_type(line)
            if stmttype is None:
                #print 'skip', repr(line)
                return
            if stmttype == 'INSERT':
                status = 'data'
                assert columns
                # transform INSERT INTO "cw_Revision" into INSERT INTO cw_Revision
                # since the former is case sensitive while the later isn't
                insert, table, values = line.split('"', 2)
                # add columns name, they may be in different orders in the target
                # database
                line = '%s %s(%s) %s' % (insert, table, ','.join(columns), values)
            elif stmttype in ('CREATE', 'BEGIN', 'COMMIT'):
                columns = []
                status = 'schema'
        if status == 'data':
            if stmt is None:
                stmt = line
            else:
                stmt += line
        #else:
        #    print 'skip', repr(line)
        if status == 'schema':
            for w in line.split():
                w = w.replace('(', '').replace(')', '').strip()
                # islower to filter out table names
                if (w.startswith('cw_') or w.startswith('eid_')) and w.islower():
                    columns.append(w)
                if w == 'CONSTRAINT':
                    break
        if line.endswith(');\n'):
            # ensure next line is a new statement
            try:
                stmttype = statement_type(nextline)
            except:
                return
            if stmttype is None:
                return
            status = None
            if stmt:
                #print stmt
                cu(stmt)
            stmt = None

    def statement_type(line):
        try:
            stmttype, remaining = line.split(None, 1)
        except ValueError:
            #print 'skip', repr(line)
            return
        if stmttype in ('INSERT', 'CREATE', 'BEGIN', 'COMMIT'):
            return stmttype
        raise Exception('unknown statement type %s' % stmttype)

    prevline = None
    for line in file(dumpfile):
        if prevline is None:
            prevline = line
        else:
            process_line(prevline, line)
            prevline = line
    process_line(prevline, line)

    session.system_sql("UPDATE ENTITIES SET source='system' WHERE source='%s'"
                       % vcssource.uri)
    session.commit()

# now inline from_repository
sync_schema_props_perms('from_repository', syncperms=False)

session.set_pool()
fpath = join(dirname(__file__), '..', 'schema', '_regproc.sql.postgres')
sqlexec(file(fpath).read(), session.system_sql, withpb=False, delimiter=';;')
session.commit()

# remove vcssource from sources file
sources = config.read_sources_file()
del sources[vcssource.uri]
config.write_sources_file(sources)
# cleanup repo sources
del repo.sources[-1]
del repo.sources_by_uri[vcssource.uri]

# fix parent_revision relation of svn repo, broken in earlier version for
# revistion created through cw
for repoeid, in rql('Any R WHERE R is Repository, R type "subversion"',
                    ask_confirm=False):
    sql('INSERT INTO parent_revision_relation(eid_from, eid_to) '
        'SELECT R1.cw_eid, R2.cw_eid FROM cw_Revision as R1, cw_Revision as R2 '
        'WHERE NOT EXISTS(SELECT 1 FROM parent_revision_relation as p WHERE R1.cw_eid=p.eid_from AND R2.cw_eid=p.eid_to) '
        'AND R2.cw_revision = R1.cw_revision - 1 '
        'AND R1.cw_from_repository=%s AND R2.cw_from_repository=%s'
        % (repoeid, repoeid), ask_confirm=False)

from cubes.vcsfile.bridge import set_at_revision

rset = rql('Revision X ORDERBY R WHERE X revision R', ask_confirm=False)
for reid, in rset:
    set_at_revision(session, reid, safetybelt=True)

checkpoint()



"""primary views for entity types defined by the vcsfile package

:organization: Logilab
:copyright: 2007-2009 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""
__docformat__ = "restructuredtext en"
_ = unicode

from logilab.mtconverter import TransformError, xml_escape

from cubicweb.selectors import implements
from cubicweb.view import EntityView
from cubicweb.common import mttransforms, tags
from cubicweb.web import uicfg
from cubicweb.web.views import (ibreadcrumbs, idownloadable,
                                baseviews, primary, tabs)

_pvs = uicfg.primaryview_section
_pvs.tag_attribute(('VersionedFile', 'name'), 'hidden')
_pvs.tag_attribute(('VersionedFile', 'directory'), 'hidden')
_pvs.tag_subject_of(('*', 'ancestor_revision', '*'), 'hidden')
_pvs.tag_object_of(('*', 'from_repository', '*'), 'hidden')
for etype in ('DeletedVersionContent', 'VersionContent'):
    _pvs.tag_attribute((etype, 'revision'), 'hidden')
    _pvs.tag_attribute((etype, 'author'), 'hidden')
    _pvs.tag_subject_of((etype, 'content_for', '*'), 'hidden')


def render_entity_summary(self, entity):
    if entity.description:
        self.w(tags.div(entity.description, klass='summary'))
    if not entity.is_head(entity.rev.branch):
        msg = self.req._('this file has newer revisions')
        self.w(tags.div(msg, klass='warning'))


class RepositoryPrimaryView(tabs.TabsMixin, primary.PrimaryView):
    __select__ = implements('Repository')

    tabs = [_('repositoryinfo_tab'), _('repositoryfiles_tab'), _('repositoryhistory_tab')]
    default_tab = 'repositoryinfo_tab'

    def cell_call(self, row, col):
        entity = self.complete_entity(row, col)
        self.render_entity_title(entity)
        self.render_tabs(self.tabs, self.default_tab, entity)


class RepositoryInfoTab(primary.PrimaryView):
    id = 'repositoryinfo_tab'
    __select__ = implements('Repository')

    def render_entity_title(self, entity):
        pass

    def render_entity_attributes(self, entity):
        super(RepositoryInfoTab, self).render_entity_attributes(entity)
        rset = self.req.execute('Any REV, REVB, REVN ORDERBY REVN DESC '
                                'WHERE REV branch REVB, REV revision REVN, REV from_repository R, '
                                'R eid %(r)s, NOT X parent_revision REV',
                                {'r': entity.eid}, 'r')
        if rset:
            self.w('<h3>%s</h3>' % self.req._('heads'))
            self.wview('table', rset)


class RepositoryFilesTab(EntityView):
    id = 'repositoryfiles_tab'
    __select__ = implements('Repository')

    def cell_call(self, row, col):
        entity = self.rset.get_entity(row, col)
        filesrql = entity.related_rql('from_repository', 'object', ('VersionedFile',))
        self.wview('list', self.req.execute(filesrql, {'x': entity.eid}, 'x'), 'noresult')


class RepositoryHistoryTab(EntityView):
    id = 'repositoryhistory_tab'
    __select__ = implements('Repository')

    def cell_call(self, row, col):
        entity = self.rset.get_entity(row, col)
        revisionsrql = entity.related_rql('from_repository', 'object', ('Revision',))
        self.wview('table', self.req.execute(revisionsrql, {'x': entity.eid}, 'x'), 'noresult')


class DVCPrimaryView(primary.PrimaryView):
    __select__ = implements('DeletedVersionContent')

    def render_entity_title(self, entity):
        title = self.req._('Revision %(revision)s of %(file)s: DELETED') % {
            'revision': entity.revision, 'file': entity.file.view('oneline')}
        self.w('<h1>%s</h1>' % title)

    render_entity_summary = render_entity_summary


class VCPrimaryView(idownloadable.IDownloadablePrimaryView):
    __select__ = implements('VersionContent')

    def render_entity_title(self, entity):
        title = self.req._('Revision %(revision)s of %(file)s') % {
            'revision': entity.revision, 'file': entity.file.view('oneline')}
        self.w('<h1>%s</h1>' % title)

    render_entity_summary = render_entity_summary


class VCMetaDataView(baseviews.MetaDataView):
    """paragraph view of some metadata"""
    __select__ = implements('VersionContent', 'DeletedVersionContent')

    def cell_call(self, row, col):
        _ = self.req._
        entity = self.entity(row, col)
        self.w(u'<div class="metadata">')
        self.w(u'#%s - ' % entity.eid)
        self.w(u'<span>%s</span> ' % _('revision %s of') % entity.revision)
        self.w(u'<span class="value">%s</span>,&nbsp;'
               % xml_escape(entity.file.path))
        self.w(u'<span>%s</span> ' % _('created on'))
        self.w(u'<span class="value">%s</span>'
               % self.format_date(entity.creation_date))
        if entity.author:
            self.w(u'&nbsp;<span>%s</span> ' % _('by'))
            self.w(tags.span(entity.author, klass='value'))
        self.w(u'</div>')


class VFPrimaryView(primary.PrimaryView):
    __select__ = implements('VersionedFile')

    def render_entity_attributes(self, entity):
        super(VFPrimaryView, self).render_entity_attributes(entity)
        self.w(u'<div class="content">')
        if entity.deleted_in_branch():
            self.w(_('this file is currently deleted in the version control system'))
        else:
            head = entity.head
            contenttype = head.download_content_type()
            if contenttype.startswith('image/'):
                self.wview('image', head.rset, row=head.row)
            else:
                try:
                    if mttransforms.ENGINE.has_input(contenttype):
                        self.w(head.printable_value('data'))
                except TransformError:
                    pass
                except Exception, ex:
                    msg = self.req._("can't display data, unexpected error: %s") % ex
                    self.w('<div class="error">%s</div>' % xml_escape(msg))
        self.w(u'</div>')

    def render_entity_relations(self, entity):
        revisions = entity.revisions
        self.w(u'<table class="listing">\n')
        self.w(u'<tr><th>%s</th><th>%s</th><th>%s</th><th>&nbsp;</th></tr>\n' %
               (_('revision'), _('author'), _('comment')))
        for vc in reversed(revisions):
            self.w(u'<tr><td><a href="%s">%s</a></td><td>%s</td><td>%s</td></tr>\n'
                   % (xml_escape(vc.absolute_url()), vc.revision,
                      xml_escape(vc.author or u''),
                      xml_escape(vc.description or u'')))
        self.w(u'</table>\n')


class RevisionPrimaryView(primary.PrimaryView):
    __select__ = implements('Revision')

    def render_entity_attributes(self, rev):
        self.field('description', xml_escape(rev.description))
        self.field('author', xml_escape(rev.author))
        if rev.branch:
            self.field('branch', xml_escape(rev.branch))
        self.field('from_repository', rev.repository.dc_title())
        parents = rev.related('parent_revision')
        if parents:
            self.field('parent_revision', self.view('csv', parents))
        children = rev.related('parent_revision', 'object')
        if children:
            self.field('parent_revision_object', self.view('csv', children))

    def render_entity_relations(self, rev):
        versioned = rev.reverse_from_revision
        if versioned:
            self.w(u'<table class="listing">')
            self.w(u'<tr><th>%s</th></tr>' %
                   self.req._('objects concerned by this revision'))
            for obj in versioned:
                self.w(u'<tr><td><a href="%s">%s</a></td></tr>' % (
                    obj.absolute_url(), xml_escape(obj.dc_title())))
            self.w(u'</table>')
        else:
            self.w(u'<p>%s</p>' % self.req._('this revision hasn\'t modified any files'))



class DVCBreadCrumbView(ibreadcrumbs.BreadCrumbView):
    id = 'breadcrumbtext'
    __select__ = implements('VersionContent', 'DeletedVersionContent')

    def cell_call(self, row, col):
        entity = self.rset.get_entity(row, col)
        self.w(entity.rev.dc_title())



# register VERSION_DATA stored procedure #######################################

from rql.utils import register_function, FunctionDescr

class version_data(FunctionDescr):
    supported_backends = ('postgres', 'sqlite',)
    rtype = 'Bytes'

try:
    register_function(version_data)
except AssertionError:
    pass


# repository specific stuff ####################################################

try:
    from cubicweb.server import SQL_CONNECT_HOOKS
except ImportError: # no server installation
    pass
else:

    options = (
        ('check-revision-interval',
         {'type' : 'int',
          'default': 5*60,
          'help': 'interval between checking of new revisions in repositories \
(default to 5 minutes).',
          'inputlevel': 2,
          'group': 'vcsfile',
          }),
        ('check-revision-commmit-every',
         {'type' : 'int',
          'default': 1,
          'help': 'after how much new imported revisions the transaction \
should be commited (default to 1, e.g. on each revision).',
          'inputlevel': 2,
          'group': 'vcsfile',
          }),
        )


    # VERSION_DATA stored procedure support w/ sqlite ##########################

    def init_sqlite_connection(cnx):
        from cubes.vcsfile import bridge
        def version_data(vceid, cnx=cnx, bridge=bridge):
            try:
                cu = cnx.cursor()
                cu.execute('SELECT REV.cw_revision, REV.cw_from_repository,'
                           ' VF.cw_directory, VF.cw_name '
                           'FROM cw_VersionContent as VC,'
                           ' cw_VersionedFile as VF, cw_Revision as REV '
                           'WHERE VC.cw_eid=%(eid)s AND'
                           ' VC.cw_content_for=VF.cw_eid AND'
                           ' VC.cw_from_revision=REV.cw_eid', {'eid': vceid})
                rev, repoeid, directory, fname = cu.fetchone()
                repohdlr = bridge.repository_handler(repoeid)
                return buffer(repohdlr.file_content(directory, fname, rev))
            except:
                import traceback
                traceback.print_exc()
                raise
        cnx.create_function('version_data', 1, version_data)

    sqlite_hooks = SQL_CONNECT_HOOKS.setdefault('sqlite', [])
    sqlite_hooks.append(init_sqlite_connection)


    # necesary posgres plpython initialization #################################

    def init_postgres_connection(cnx):
        # we'll have to initialize plpython sys.path since cw customize
        # it, as well as cubes.__path__, so give information using dedicated
        # sys_path_init registered procedure (in schema/_reproc.sql.postgres)
        import sys
        from os.path import join
        from cubicweb.cwconfig import CubicWebNoAppConfiguration
        # XXX enhance filtering
        stdlibdir = join(sys.prefix, 'lib')
        exportpath = [p for p in sys.path if not p.startswith(stdlibdir)]
        cubespath = CubicWebNoAppConfiguration.cubes_search_path()
        cu = cnx.cursor()
        # use '||'.join(exportpath) since plpython doesn't support sequence as
        # argument
        try:
            cu.execute('SELECT sys_path_init(%(syspath)s, %(cubespath)s)',
                       {'syspath': '||'.join(exportpath),
                        'cubespath': '||'.join(cubespath)})
            cnx.commit()
        except Exception, ex:
            print ('error while calling sys_path_init for posgtgres plpython '
                   'initialization: %s' % ex)
            print ("don't worry if you're creating the instance or migrating "
                   "vcsfile to 1.0")
            cnx.rollback()

    postgres_hooks = SQL_CONNECT_HOOKS.setdefault('postgres', [])
    postgres_hooks.append(init_postgres_connection)



"""cubicweb-vcsfile"""

try:
    from cubicweb import server
except ImportError:
    # no server part
    pass
else:
    from logilab.common.modutils import LazyObject
    server.SOURCE_TYPES['vcsfile'] = LazyObject('cubes.vcsfile.vcssource',
                                                'VCSSource')

from cubicweb import ETYPE_NAME_MAP
ETYPE_NAME_MAP['Versionedfile'] = 'VersionedFile'
ETYPE_NAME_MAP['Deletedversioncontent'] = 'DeletedVersionContent'
ETYPE_NAME_MAP['Versioncontent'] = 'VersionContent'


IMMUTABLE_ATTRIBUTES = frozenset(('VersionedFile.directory',
                                  'VersionedFile.name',
                                  'VersionedFile.from_repository',
                                  'Revision.revision',
                                  'Revision.changeset',
                                  'Revision.branch',
                                  'Revision.from_repository',
                                  'DeletedVersionContent.from_revision',
                                  'DeletedVersionContent.content_for',
                                  'VersionContent.from_revision',
                                  'VersionContent.content_for',
                                  'VersionContent.data',
                                  ))



"""entity types used to represent a version control system (vcs) content

:organization: Logilab
:copyright: 2007-2010 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""
__docformat__ = "restructuredtext en"
_ = unicode

from yams.buildobjs import (EntityType, RelationType, SubjectRelation,
                            String, Int, Bytes)


class Repository(EntityType):
    """a local subversion repository which will be used as versionned content
    entity source
    """
    permissions = {
        'read':   ('managers', 'users'),
        'add':    ('managers',),
        'update': ('managers',),
        'delete': ('managers',),
        }
    type = String(required=True, vocabulary=('subversion', 'mercurial'),
                  description=_('repository\'s type'))
    path = String(required=True, unique=True,
                  description=_('path to the repository'))
    subpath = String(description=_('if specified, only import the given subpart'
                                   ' of the repository (path relative to the '
                                   'repository root)'))
    encoding = String(default='utf-8', maxsize=20, required=True,
                      description=_('encoding used for the repository (e.g.'
                                    ' for file path and check-in comments)'))


class Revision(EntityType):
    permissions = {
        'read':   ('managers', 'users', 'guests'),
        'add':    ('managers', 'users',),
        'update': ('managers', 'owners'),
        'delete': (),
        }
    # NOTE: creation_date will be set at revision date
    revision = Int(required=True, indexed=True,
                   description=_('revision number.'))

    description = String(fulltextindexed=True, description=_('comment for this revision.'))

    author = String(indexed=True, fulltextindexed=True, maxsize=200,
                    description=_("author of this revision."))

    changeset = String(indexed=True, maxsize=100,
                       description=_('change set identifier, when used by the '
                                     'underlying version control system.'))

    branch = String(indexed=True, maxsize=255,
                    description=_("branch of this revision."))
    # XXX tags

    from_repository = SubjectRelation('Repository', cardinality='1*', composite='object')
    parent_revision = SubjectRelation('Revision')
    ancestor_revision = SubjectRelation('Revision')


class VersionedFile(EntityType):
    """a file stored in a versioned source"""
    permissions = {
        'read':   ('managers', 'users', 'guests'),
        'add':    ('managers', 'users',),
        'update': (),
        'delete': (),
        }
    directory = String(required=True, fulltextindexed=True, indexed=True,
                       description=_('directory of the file in the repository.'))
    name = String(required=True, fulltextindexed=True,
                  description=_('name of the file in the repository.'))
    from_repository = SubjectRelation('Repository', cardinality='1*', composite='object')


class DeletedVersionContent(EntityType):
    """represent a revision where a versioned file has been deleted"""
    permissions = {
        'read':   ('managers', 'users', 'guests'),
        'add':    ('managers', 'users',),
        'update': ('managers', 'owners'),
        'delete': (),
        }
    from_revision = SubjectRelation('Revision', cardinality='1*', composite='object')
    at_revision = SubjectRelation('Revision')


class VersionContent(DeletedVersionContent):
    """actual content of a versioned file, for a given revision"""
    # not required to avoid yams validation error since it's not actually stored
    # in the database
    data = Bytes(fulltextindexed=True, description=_("file's data"))
    data_format = String(required=True, meta=True, maxsize=50,
                         description=_('MIME type of the file.'))
    data_encoding = String(meta=True, maxsize=20,
                           description=_('encoding of the file when it applies (e.g. text).'))

    vc_copy = SubjectRelation('VersionContent', cardinality='?*')
    vc_rename = SubjectRelation('VersionContent', cardinality='?*')


class from_repository(RelationType):
    inlined = True

class content_for(RelationType):
    subject = ('DeletedVersionContent', 'VersionContent')
    object = 'VersionedFile'
    composite = 'object'
    cardinality = '1*'
    inlined = True

class from_revision(RelationType):
    inlined = True

class at_revision(RelationType):
    permissions = {
        'read':   ('managers', 'users', 'guests'),
        'add':    (),
        'delete': (),
        }

class vc_copy(RelationType):
    inlined = True

class vc_rename(RelationType):
    inlined = True



"""entity classes for vcsfile content types


:organization: Logilab
:copyright: 2007-2010 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""
__docformat__ = "restructuredtext en"

from logilab.common.decorators import cached, clear_cache
from logilab.common.compat import any

from cubicweb import QueryError, Binary
from cubicweb.interfaces import IPrevNext, IDownloadable
from cubicweb.entities import AnyEntity, fetch_config

from cubes.vcsfile import queries

def rql_revision_content(repoeid, revnum, branch=None):
    """return rql query to get the repository content at a given revision"""
    # XXX: MAX(V) is a trick necessary because RQL doesn't support subqueries,
    #      which would be necessary to have the correct query. This trick is ok
    #      since we know that entity with MAX(V) has MAX(VR) as well, due to
    #      source implementation
    args = {'rev': revnum, 'x': repoeid, 'branch': branch}
    return ('Any MAX(V),MAX(REV),F,D,N GROUPBY D,N,F ORDERBY D,N '
            'WHERE V content_for F, V from_revision VR, VR revision REV, '
            'VR branch %(branch)s, '
            'F directory D, F name N, F from_repository R, R eid %(x)s, '
            'F is VersionedFile, (VR revision <= %(rev)s AND NOT EXISTS('
            'X is DeletedVersionContent, X content_for F, X from_revision XR, '
            'XR branch %(branch)s, XR revision < %(rev)s, XR revision >= REV))'
            ), args

_MARKER = object()


class Repository(AnyEntity):
    """customized class for Repository entities"""
    id = 'Repository'
    fetch_attrs, fetch_order = fetch_config(['path', 'subpath', 'type'])
    rest_attr = 'eid' # see #XXX, using path cause pb w/ apache redirection

    def dc_title(self):
        title = '%s:%s' % (self.type, self.path)
        if self.subpath:
            title += ' (%s)' % self.subpath
        return title

    def clear_all_caches(self):
        super(VersionedFile, self).clear_all_caches()
        clear_cache(self, 'branch_head')

    # navigation in versioned content #########################################

    def latest_known_revision(self):
        """return the number of the latest known revision"""
        latestrev = self.req.execute(
            'Any MAX(REVR) WHERE R eid %(r)s, REV from_repository R, '
            'REV revision REVR', {'r' : self.eid}, 'r')[0][0]
        if latestrev is None:
            return -1
        return latestrev

    def branches(self):
        """return existing branches"""
        return [b for b, in self.req.execute(
            'DISTINCT Any B WHERE R eid %(r)s, REV from_repository R, '
            'REV branch B', {'r' : self.eid}, 'r')]

    @cached
    def branch_head(self, branch=_MARKER):
        """return latest revision of the given branch"""
        if branch is _MARKER:
            branch = self.default_branch()
        rset = self.req.execute(
            'Any MAX(REV) WHERE V at_revision REV, REV branch %(branch)s, '
            'REV from_repository R, R eid %(r)s',
            {'r': self.eid, 'branch': branch}, 'r')
        if rset[0][0] is None:
            return None
        return rset.get_entity(0, 0)

    def versioned_file(self, directory, filename):
        rset = self.req.execute(
            'Any X WHERE X is VersionedFile, X from_repository R, '
            'R eid %(repo)s, X directory %(dir)s, X name %(name)s',
            {'repo' : self.eid, 'dir' : directory, 'name' : filename})
        return rset and rset.get_entity(0, 0) or None

    def is_directory_deleted(self, directory):
        # XXX same MAX(X) trick as above
        rset = self.req.execute(
            'Any MAX(VC), MAX(REV) GROUPBY VF '
            'WHERE VC from_revision R, R revision REV, R from_repository X, '
            'X eid %(x)s, VC content_for VF, VF directory ~= %(dir)s',
            {'x': self.eid, 'dir': directory}, 'x')
        return not any(e for e in rset.entities() if e.id == 'VersionContent')

    # vcs write support ########################################################

    def default_branch(self):
        return {'subversion': None, 'mercurial': u'default'}[self.type]

    def make_revision(self, msg=None, author=_MARKER, branch=_MARKER,
                      parentbranch=_MARKER):
        """parameters:
        :msg: the check-in message (string)
        :author: the check-in author (string)
        :branch: the branch in which the revision should be done (string)
        :parentbranch: when a branch is created, the parent branch of the revision
        """
        if branch is _MARKER:
            branch = self.default_branch()
        if author is _MARKER:
            author = self.req.user.login
        parent = self.branch_head(branch)
        if parent is None and branch != parentbranch:
            parent = self.branch_head(parentbranch)
        #assert parent
        args = {'description' : msg, 'author': author, 'branch': branch,
                'repoeid' : self.eid, 'parent': parent and parent.eid}
        rset = self.req.execute(queries.new_revision_rql(parent), args)
        return rset.get_entity(0, 0)

    def vcs_add(self, filedir, filename, stream, strict=False,
                extraattrs=None, **kwargs):
        """add the file into the repository, as <filedir>/<filename>, with
        content specified in <stream> (Binary instance).

        If strict and the file already exists in the repository, an error will
        be raised, else a new revision will be simply uploaded.

        If you've an extended versioned file's schema you can give additional
        attributes to create_entity using the extraattrs argument (expected to
        be a dict)

        Return the VersionedFile instance.
        """
        vf = self.versioned_file(filedir, filename)
        if not vf:
            if extraattrs is None:
                extraattrs = {}
            vf = self.req.create_entity('VersionedFile',
                                        from_repository=self, directory=filedir,
                                        name=filename, **extraattrs)
        elif strict:
            raise Exception('attempting to add an already existing file')
        vf.vcs_upload_revision(stream, **kwargs)
        return vf


class Revision(AnyEntity):
    """customized class for Revision entities"""
    id = 'Revision'
    fetch_attrs, fetch_order = fetch_config(['revision', 'changeset', 'author',
                                             'description', 'creation_date'])
    __implements__ = AnyEntity.__implements__ + (IPrevNext,)

    def dc_title(self):
        if self.changeset:
            return self.req._('revision #%(rev)s:%(changeset)s') % {
                'rev': self.revision, 'changeset': self.changeset}
        else:
            return self.req._('revision #%s') % self.revision

    def dc_long_title(self):
        return self.req._('%(rev)s of repository %(repo)s') % {
            'rev': self.dc_title(), 'repo': self.repository.dc_title()}

    def sortvalue(self, rtype=None):
        if rtype is None:
            return self.revision
        return super(Revision, self).sortvalue(rtype)

    def previous_entity(self): # IPrevNext
        # may have multiple parents, take the first one in the same branch...
        rset = self.req.execute('Any X LIMIT 1 WHERE R parent_revision X , '
                                'R eid %(r)s, X branch %(branch)s',
                                {'r': self.eid, 'branch': self.branch})
        if rset:
            return rset.get_entity(0, 0)

    def next_entity(self): # IPrevNext
        # may have multiple children, take the first one in the same branch...
        rset = self.req.execute('Any X LIMIT 1 WHERE X parent_revision R , '
                                'R eid %(r)s, X branch %(branch)s',
                                {'r': self.eid, 'branch': self.branch})
        if rset:
            return rset.get_entity(0, 0)

    def parent(self): # for breadcrumbs
        return self.repository

    def pre_web_edit(self):
        """callback called by the web editcontroller when an entity will be
        created/modified, to let a chance to do some entity specific stuff.
        """
        if not self.has_eid():
            try:
                repoeid = self.linked_to('from_repository', 'subject', remove=False)[0]
            except IndexError:
                raise RequestError('missing repository information')
            self.req.set_shared_data('vcsrepoeid', repoeid, querydata=True)

    # navigation in versioned content #########################################

    @property
    def repository(self):
        return self.from_repository[0]


class VersionedFile(AnyEntity):
    """customized class for VersionedFile entities"""
    id = 'VersionedFile'
    fetch_attrs, fetch_order = fetch_config(['directory', 'name'])

    # XXX branches
    def dc_title(self):
        if self.deleted_in_branch():
            return '%s (%s)' % (self.path, self.req._('DELETED'))
        return self.path

    def dc_long_title(self):
        return self.req._('%(path)s (from repository %(repotype)s:%(repo)s)') % {
            'path': self.dc_title(), 'repo': self.repository.path,
            'repotype': self.repository.type}

    def parent(self): # for breadcrumbs
        return self.repository

    def clear_all_caches(self):
        super(VersionedFile, self).clear_all_caches()
        clear_cache(self, 'version_content')
        clear_cache(self, 'branch_head')

    # navigation in versioned content #########################################

    @property
    def repository(self):
        return self.from_repository[0]

    @property
    def path(self):
        if self.directory:
            return '%s/%s' % (self.directory, self.name)
        return self.name

    @property
    def revisions(self):
        """return an ordered list of revision contents for this file"""
        return self.reverse_content_for

    @property
    def head(self):
        return self.branch_head()

    @cached
    def version_content(self, revnum):
        if revnum is None:
            return self.branch_head()
        rset = self.req.execute(
            'Any C,VF,R WHERE C content_for VF, C from_revision R, '
            'R revision %(rnum)s, VF eid %(x)s',
            {'rnum' : int(revnum), 'x' : self.eid}, 'x')
        if rset:
            return rset.get_entity(0, 0)
        return None

    @cached
    def branch_head(self, branch=_MARKER):
        """return latest [deleted] version content for this file in the given
        branch
        """
        if branch is _MARKER:
            branch = self.repository.default_branch()
        rset = self.req.execute(
            'Any MAX(V) WHERE X eid %(x)s, V content_for X, V at_revision REV, '
            'REV branch %(branch)s',
            {'x': self.eid, 'branch': branch}, 'x')
        # content_for relation is not mandatory on VersionedFile entities
        if rset[0][0] is None:
            return None
        return rset.get_entity(0, 0)

    def deleted_in_branch(self, branch=_MARKER):
        """deleted in branch != does not even exist in branch
        """
        head = self.branch_head(branch)
        return head and head.is_deletion() or False

    # vcs write support #######################################################

    def vcs_rm(self, rev=None, **kwargs):
        """remove this file from the vcs

        takes a rev or a kwargs having keys: branch, msg, author
        """
        if not isinstance(self.branch_head(kwargs.get('branch', _MARKER)),
                          VersionContent):
            # already deleted
            raise QueryError(self.req._('%s is already deleted from the vcs')
                             % self.path)
        if rev is None:
            if not 'msg' in kwargs:
                kwargs['msg'] = self.req._('deleted %s') % self.dc_title()
            rev = self.repository.make_revision(**kwargs)
        self.req.execute(
            'INSERT DeletedVersionContent D: D content_for X, '
            'D from_revision R WHERE X eid %(x)s, R eid %(r)s',
            {'x' : self.eid, 'r' : rev.eid}, ('x', 'r'))

    def vcs_upload_revision(self, stream, rev=None, **kwargs):
        """upload a new revision for this file to the vcs

        either takes a rev or a kwargs having keys: branch, msg, author

        return the newly created version content entity
        """
        if rev is None:
            rev = self.repository.make_revision(**kwargs)
        return self.req.execute(queries.new_file_content_rql(),
                                {'vfeid' : self.eid, 'reveid' : rev.eid,
                                 'data' : Binary(stream.read())}).get_entity(0, 0)


class DeletedVersionContent(AnyEntity):
    """customized class for DeletedVersionContent entities"""
    id = 'DeletedVersionContent'
    fetch_attrs, fetch_order = fetch_config(['from_revision', 'content_for'], order='DESC')
    __implements__ = AnyEntity.__implements__ + (IPrevNext,)

    def dc_title(self):
        return self.req._('%(file)s DELETED (revision %(revision)s)') % {
            'file': self.file.path, 'revision': self.rev.revision}

    def dc_long_title(self):
        rev = self.rev
        return self.req._('%(file)s DELETED (revision %(revision)s on %(date)s by %(author)s)') % {
            'file': self.file.path, 'revision': rev.revision,
            'author': rev.author, 'date': rev.printable_value('creation_date')}

    def dc_description(self, format='text/plain'):
        # override cubicweb's default implementation because it requires
        # 'description' to be a real schema attribute but it's only
        # a class property in our case
        return self.rev.dc_description(format)

    def previous_entity(self): # IPrevNext
        rset = self.req.execute('Any PV,R,PVR,VF ORDERBY PVR DESC LIMIT 1 '
                                'WHERE PV from_revision R, R revision PVR, '
                                'R branch %(b)s, PV content_for VF, '
                                'R revision < %(r)s, VF eid %(vf)s',
                                {'r': self.rev.revision, 'b': self.rev.branch,
                                 'vf': self.file.eid})
        if rset:
            return rset.get_entity(0, 0)

    def next_entity(self): # IPrevNext
        rset = self.req.execute('Any NV,R,NVR,VF ORDERBY NVR ASC LIMIT 1 '
                                'WHERE NV from_revision R, R revision NVR, '
                                'R branch %(b)s, NV content_for VF, '
                                'R revision > %(r)s, VF eid %(vf)s',
                                {'r': self.rev.revision, 'b': self.rev.branch,
                                 'vf': self.file.eid})
        if rset:
            return rset.get_entity(0, 0)

    def parent(self): # for breadcrumbs
        return self.file

    # navigation in versioned content #########################################

    @property
    def repository(self):
        return self.rev.repository

    @property
    def file(self):
        return self.content_for[0]

    @property
    def rev(self):
        return self.from_revision[0]

    @property
    def head(self):
        return self.file.branch_head(self.rev.branch)

    # < 0.7 bw compat properties
    @property
    def revision(self):
        return self.from_revision[0].revision

    @property
    def author(self):
        return self.from_revision[0].author

    @property
    def description(self):
        return self.from_revision[0].description

    def is_head(self, branch=_MARKER):
        """return true if this version content is the head for its file in
        the given branch or in its revision's branch.
        """
        if branch is _MARKER:
            branch = self.rev.branch
        return self.eid == self.file.branch_head(branch).eid

    def is_deletion(self):
        return True


class VersionContent(DeletedVersionContent):
    """customized class for VersionContent entities"""
    id = 'VersionContent'
    __implements__ = DeletedVersionContent.__implements__ + (IDownloadable,)

    def dc_title(self):
        return self.req._('%(file)s (revision %(revision)s)') % {
            'file': self.file.path, 'revision': self.revision}

    def dc_long_title(self):
        rev = self.rev
        return self.req._('%(file)s (revision %(revision)s on %(date)s by %(author)s)') % {
            'file': self.file.path, 'revision': rev.revision,
            'author': rev.author, 'date': rev.printable_value('creation_date')}

    def posted_data(self, fetch=True):
        """use this method to get VersionContent.data in
        [before|after]_add_entity hooks
        """
        try:
            return self.__data
        except:
            data = self.pop('data', None)
            if data is None and fetch:
                data = self.data
            self.__data = data
            return data

    def is_deletion(self):
        return False

    # IDownloadable
    def download_url(self):
        return self.absolute_url(vid='download')
    def download_content_type(self):
        return self.data_format
    def download_encoding(self):
        return self.data_encoding
    def download_file_name(self):
        return self.file.name
    def download_data(self):
        return self.data.getvalue()

    def size(self):
        rql = "Any LENGTH(D) WHERE X eid %(eid)s, X data D"
        return self.req.execute(rql, {'eid': self.eid})[0][0]



"""custom forms to submit new revision to the svn repository or to edit
some information about existing revisions.

:organization: Logilab
:copyright: 2007-2010 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""
__docformat__ = "restructuredtext en"
_ = unicode

from copy import copy

from logilab.mtconverter import xml_escape
from logilab.common.decorators import cached

from cubicweb.selectors import (match_form_params, match_kwargs,
                                implements, entity_implements,
                                specified_etype_implements)
from cubicweb.view import EntityView
from cubicweb.common import tags
from cubicweb.web import (Redirect, controller, form,
                          formfields as fields, formwidgets as fwdgs)
from cubicweb.web.views import autoform, editforms, forms

from cubes.vcsfile import IMMUTABLE_ATTRIBUTES


class RevisionCreationForm(autoform.AutomaticEntityForm):
    # vfeid needed for the inline-creation of VersionContent
    __select__ = specified_etype_implements('Revision') & match_form_params('vfeid')

    def should_display_inline_creation_form(self, rschema, existant, card):
        if rschema == 'from_revision':
            return True
        return super(RevisionCreationForm, self).should_display_inline_creation_form(
            rschema, existant, card)

    def default_author(self):
        return self.req.user.login


class VCInlinedCreationFormView(editforms.InlineEntityCreationFormView):
    # vfeid needed for the inline-creation of VersionContent
    __select__ = (match_form_params('vfeid') & match_kwargs('peid', 'rtype')
                  & specified_etype_implements('VersionContent'))

    if hasattr(editforms.InlineEntityCreationFormView, 'form'):

        # cw > 3.5.2
        @cached
        def _entity(self):
            """creation view for an entity"""
            vf = self.req.entity_from_eid(int(self.req.form['vfeid']))
            self.orig_entity = vf.head
            # need a fully completed entity before doing the copy
            self.orig_entity.complete()
            if self.orig_entity.attr_metadata('data', 'format') in (
                'text/plain', 'text/html', 'text/rest'):
                # fill cache, Bytes fields are not loaded by complete()
                self.orig_entity.data
            entity = copy(self.orig_entity)
            self.initialize_varmaker()
            entity.eid = self.varmaker.next()
            return entity

        def add_hiddens(self, form, entity):
            super(VCInlinedCreationFormView, self).add_hiddens(
                form, entity)
            form.form_add_hidden('content_for', self.orig_entity.file.eid,
                                 eidparam=True)

    else:

        def call(self, etype, peid, rtype, role='subject', **kwargs):
            """creation view for an entity"""
            vf = self.req.entity_from_eid(int(self.req.form['vfeid']))
            self.orig_entity = vf.head
            # need a fully completed entity before doing the copy
            self.orig_entity.complete()
            if self.orig_entity.attr_metadata('data', 'format') in (
                'text/plain', 'text/html', 'text/rest'):
                # fill cache, Bytes fields are not loaded by complete()
                self.orig_entity.data
            entity = copy(self.orig_entity)
            self.initialize_varmaker()
            entity.eid = self.varmaker.next()
            self.render_form(entity, peid, rtype, role)

        def add_hiddens(self, form, entity, peid, rtype, role):
            super(VCInlinedCreationFormView, self).add_hiddens(
                form, entity, peid, rtype, role)
            form.form_add_hidden('content_for', self.orig_entity.file.eid,
                                 eidparam=True)

def filter_out_immutable_rels(func):
    def wrapper(self, *args, **kwargs):
        entity = self.edited_entity
        if entity.has_eid():
            immutable = IMMUTABLE_ATTRIBUTES
        else:
            immutable = ('Revision.revision',)
        for rschema, tschemas, role in func(self, *args, **kwargs):
            if role == 'subject':
                etype = entity.id
            # take care, according to the wrapped function tschemas is a list or
            # a single schema
            elif isinstance(tschemas, list):
                etype = tschemas[0]
            else:
                etype = tschemas
            if '%s.%s' % (etype, rschema) in immutable:
                continue
            yield rschema, tschemas, role
    return wrapper


class VCRevisionEditionForm(autoform.AutomaticEntityForm):
    __select__ = implements('Revision', 'VersionedFile',
                            'VersionContent', 'DeletedVersionContent')
    if hasattr(autoform.AutomaticEntityForm, '_relations_by_section'):
        # cw 3.6
        _relations_by_section = filter_out_immutable_rels(
            autoform.AutomaticEntityForm._relations_by_section)
    else:
        relations_by_category = filter_out_immutable_rels(
            autoform.AutomaticEntityForm.relations_by_category)
        srelations_by_category = filter_out_immutable_rels(
            autoform.AutomaticEntityForm.srelations_by_category)
        inlined_relations = filter_out_immutable_rels(
            autoform.AutomaticEntityForm.inlined_relations)


def available_branches(form):
    return form.edited_entity.repository.branches()


class NewRevisionForm(forms.EntityFieldsForm):
    id = 'vfnewrevform'
    __select__ = entity_implements('VersionedFile')

    form_renderer_id = 'base'
    form_buttons = [fwdgs.SubmitButton()]

    file = fields.FileField(eidparam=True)
    branch = fields.StringField(required=True, eidparam=True,
                                widget=fwdgs.Select, choices=available_branches,
                                # XXX in 3.6, uncomment this / remove formvalues
                                #initial=lambda formentity.repository.default_branch()
                                )
    msg = fields.StringField(label=_('commit message'), eidparam=True)


class VFUploadFormView(form.FormViewMixIn, EntityView):
    """form to upload a new revision of a versioned file"""
    id = 'vfnewrevform'
    __select__ = implements('VersionedFile')

    submitmsg = _('new revision has been checked-in')
    action = 'vcsnewrev'

    def call(self, **_kwargs):
        # XXX explain why we would want self.rset.rowcount != 1
        for i in xrange(self.rset.rowcount):
            self.cell_call(i, 0)

    def form_title(self, entity):
        return self.req._('Upload a new revision of %s') % (
            '<a href="%s">%s</a>' % (xml_escape(entity.absolute_url()),
                                     xml_escape(entity.dc_title())))

    def cell_call(self, row, col, **kwargs):
        entity = self.entity(row, col)
        self.w(u'<div class="iformTitle">')
        self.w(self.form_title(entity))
        self.w(u'</div>')
        form = self.vreg['forms'].select(
            'vfnewrevform', self.req, entity=entity,
            submitmsg=self.req._(self.submitmsg), action=self.action,
            __redirectpath=entity.rest_path(),)
        self.additional_form_fields(form, entity)
        # give default values so it won't try to get them from the entity
        self.w(form.render(formvalues=dict(branch=entity.repository.default_branch(),
                                           file=u'', msg=u'')))

    def additional_form_fields(self, form, entity):
        pass


class VFUploadController(controller.Controller):
    """upload a new revision of an already existing document"""
    id = 'vcsnewrev'

    def publish(self, rset=None):
        for eid in self.req.edited_eids():
            formparams = self.req.extract_entity_params(eid, minparams=2)
            vf = self.req.entity_from_eid(eid)
            vf.complete()
            self.upload_revision(vf, formparams)
        try:
            kwargs = {'__message': self.req.form['__message']}
        except KeyError:
            kwargs = {}
        try:
            goto = self.req.build_url(self.req.form['__redirectpath'], **kwargs)
        except KeyError:
            goto = vf.absolute_url(**kwargs)
        raise Redirect(goto)

    def upload_revision(self, vf, formparams):
        # [-1] to discard file name and mime type
        stream = formparams['file'][-1]
        vf.vcs_upload_revision(stream, branch=formparams['branch'],
                               msg=formparams['msg'])
        return vf.eid


class VcsRmForm(form.FormViewMixIn, EntityView):
    id = 'vcsrmform'
    __select__ = implements('VersionedFile')

    def cell_call(self, row, col):
        entity = self.entity(row, col)
        self.w(u'<div class="iformTitle">')
        self.w(self.req._('DELETE the versioned file %(file)s') %
                    {'file': entity.view('incontext')})
        self.w(u'</div>')
        # XXX branch
        form = self.vreg.select_object('forms', 'base', self.req, self.rset,
                                       form_renderer_id='base', entity=entity,
                                       action=self.build_url('vcsrm'),
                                       form_buttons=[fwdgs.SubmitButton()])
        form.append_field(fields.StringField(name='msg', eidparam=True,
                                             label=_('commit message')))
        self.w(form.form_render(msg=''))


class VcsRmController(controller.Controller):
    """DELETE"""
    id = 'vcsrm'

    def publish(self, rset=None):
        req = self.req
        for eid in req.edited_eids():
            formparams = req.extract_entity_params(eid, minparams=2)
            vf = self.req.entity_from_eid(eid)
            vf.vcs_rm(msg=formparams['msg'])
        msg = req._('file was marked DELETED')
        raise Redirect(vf.absolute_url(__message=msg))



# -*- coding: utf-8 -*-
"""hooks for vcsfile content types

:organization: Logilab
:copyright: 2007-2010 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""
__docformat__ = "restructuredtext en"

import os

from logilab.mtconverter import need_guess, guess_mimetype_and_encoding

from cubicweb import QueryError, ValidationError, Binary
from cubicweb.server.hooksmanager import Hook
from cubicweb.server.pool import PreCommitOperation, Operation

from cubes.vcsfile import IMMUTABLE_ATTRIBUTES, bridge

def missing_relation_error(entity, rtype):
    # use __ since msgid recorded in cw, we don't want to translate it in
    # this cube
    __ = entity.req._
    msg = __('at least one relation %(rtype)s is required on %(etype)s (%(eid)s)')
    errors = {'from_repository': msg % {'rtype': __(rtype),
                                        'etype': __(entity.id),
                                        'eid': entity.eid}}
    return ValidationError(entity.eid, errors)

def _generate_vc_data(gen, linkedvar, rel):
    linkedvar.accept(gen)
    return 'VERSION_DATA(%s)' % linkedvar._q_sql

# initialization hooks #########################################################

class ServerStartupHook(Hook):
    """synchronize repository on server startup, and install attribute map
    on the system source sql generator
    """
    events = ('server_startup',)
    def call(self, repo):
        repo.system_source.map_attribute('VersionContent', 'data',
                                         _generate_vc_data)
        repo.looping_task(repo.config.get('check-revision-interval', 5*60),
                          bridge.import_content, repo,
                          repo.config.get('check-revision-commit-every', 5*60))


class AddRepositoryHook(Hook):
    """add repository eid to vcs bridge cache"""
    events = ('before_add_entity', )
    accepts = ('Repository',)

    def call(self, session, entity):
        try:
            # check and register the repository into the bridge
            bridge.cache_handler_for_repository(entity)
        except bridge.VCSException, ex:
            raise ex.to_validation_error(session._)

# internals to be able to create new vcs repo revision using rql queries #######

def _vc_prepare(entity, vcsrepoeid=None):
    try:
        vfeid = entity['content_for']
    except KeyError:
        raise missing_relation_error(entity, 'content_for')
    session = entity.req
    if vcsrepoeid is None:
        # retrieve associated VersionedFile instance
        vf = session.execute('Any X,D,N,R WHERE X name N, X directory D,'
                             'X from_repository R, X eid %(x)s',
                             {'x': vfeid}, 'x').get_entity(0, 0)
        vcsrepoeid = vf.repository.eid
    try:
        transaction = session.transaction_data['vctransactions'][vcsrepoeid]
    except KeyError:
        raise QueryError('you must create a Revision instance before '
                         'adding some content')
    CheckRevisionOp(session, entity=entity)
    return bridge.repository_handler(vcsrepoeid), transaction

def _vc_vf(entity):
    """return versioned file associated to a [Deleted]VersionContent entity"""
    if not hasattr(entity, '_vcsrepoinfo'):
        try:
            vfeid = entity['content_for']
        except KeyError, ex:
            raise missing_relation_error(entity, 'content_for')
        session = entity.req
        vf = session.execute(
            'Any X, R, FD, FN WHERE X directory FD, X name FN, X eid %(x)s, '
            'X from_repository R', {'x': vfeid}, 'x').get_entity(0, 0)
        entity._vcsrepo_info = vf
    return entity._vcsrepo_info


class VCTransactionOp(Operation):

    def precommit_event(self):
        bridge.set_at_revision(self.session, self.revision.eid)

    def revertprecommit_event(self):
        transactions = self.session.transaction_data.setdefault('vctransactions', {})
        for transaction in transactions.itervalues():
            transaction.rollback()

    def commit_event(self):
        transactions = self.session.transaction_data.setdefault('vctransactions', {})
        for transaction in transactions.itervalues():
            transaction.commit()

    revertcommit_event = revertprecommit_event

    def rollback_event(self):
        transactions = self.session.transaction_data.setdefault('vctransactions', {})
        for transaction in transactions.itervalues():
            transaction.rollback()


class AddRevisionHook(Hook):
    events = ('before_add_entity',)
    accepts = ('Revision',)

    def call(self, session, entity):
        if not entity.get('revision'):
            # new revision to be created, set a temporary value
            entity['revision'] = 0
        # skip further processing if the revision is being imported from the
        # vcs repository
        if session.is_internal_session:
            return
        try:
            vcsrepoeid = entity['from_repository']
        except KeyError:
            vcsrepoeid = session.transaction_data.pop('vcsrepoeid', None)
            if vcsrepoeid is None:
                raise missing_relation_error(entity, 'from_repository')
        revision = entity.get('revision', 0)
        if revision > 0: # set to 0 by hook
            raise QueryError("can't specify revision")
        transactions = session.transaction_data.setdefault('vctransactions', {})
        # should not have multiple transaction on the same repository
        if vcsrepoeid in transactions:
            raise QueryError('already processing a new revision')
        vcsrepohdlr = bridge.cache_handler_for_repository(
            session.entity_from_eid(vcsrepoeid))
        transaction = vcsrepohdlr.revision_transaction(session, entity)
        transaction.reveid = entity.eid
        transactions[vcsrepoeid] = transaction
        entity['revision'] = transaction.rev
        VCTransactionOp(session, revision=entity)


class AddVersionContentHook(Hook):
    events = ('before_add_entity',)
    accepts = ('VersionContent',)

    def call(self, session, entity):
        data = self.get_data(entity)
        # save data for usage in after_add_entity hook below
        session.transaction_data[(entity.eid, 'data')] = data
        if need_guess(entity.get('data_format'), entity.get('data_encoding')):
            vf = _vc_vf(entity)
            encoding = bridge.repository_handler(vf.repository.eid).encoding
            mt, enc = guess_mimetype_and_encoding(data=data, filename=vf.name,
                                                  fallbackencoding=encoding)
            if mt and not entity.get('data_format'):
                entity['data_format'] = unicode(mt)
            if enc and not entity.get('data_encoding'):
                entity['data_encoding'] = unicode(enc)
        # skip further processing if the revision is being imported from the
        # vcs repository
        if session.is_internal_session:
            return
        vf = _vc_vf(entity)
        vcsrepohdlr, transaction = _vc_prepare(entity, vf.repository.eid)
        vcsrepohdlr.add_versioned_file_content(session, transaction, vf, entity,
                                               data)

    def get_data(self, entity):
        session = entity.req
        if session.is_internal_session:
            vf = _vc_vf(entity)
            vcsrepohdlr = bridge.repository_handler(vf.repository.eid)
            return Binary(vcsrepohdlr.file_content(
                vf.directory, vf.name, vcsrepohdlr.imported_revision))
        # data not stored in the local database, pop it
        data = entity.posted_data(fetch=False)
        if not data:
            # missing data
            raise ValidationError(entity, {'data': session.__('required attribute')})
        return data


class AfterAddVersionContentHook(Hook):
    events = ('after_add_entity',)
    accepts = ('VersionContent',)
    def call(self, session, entity):
        # restore data for full text indexation
        entity['data'] = session.transaction_data.pop((entity.eid, 'data'))


class AddDeletedVersionContentHook(Hook):
    events = ('before_add_entity',)
    accepts = ('DeletedVersionContent',)

    def call(self, session, entity):
        # skip further processing if the revision is being imported from the
        # vcs repository
        if session.is_internal_session:
            return
        vcsrepohdlr, transaction = _vc_prepare(entity)
        vf = _vc_vf(entity)
        vcsrepohdlr.add_versioned_file_deleted_content(session, transaction, vf,
                                                       entity)


class AddVersionedFileHook(Hook):
    events = ('before_add_entity',)
    accepts = ('VersionedFile',)

    def call(self, session, entity):
        # skip further processing if the revision is being imported from the
        # vcs repository
        if session.is_internal_session:
            return
        CheckVersionedFileOp(session, entity=entity)

# ancestor_revision synchro ####################################################

class AddParentRevisionHook(Hook):
    events = ('after_add_relation',)
    accepts = ('parent_revision',)

    def call(self, session, fromeid, rtype, toeid):
        # ancestor relation may already be set
        # if session.unsafe_execute('Any X WHERE X eid %(x)s, Y eid %(y)s, '
        #                           'X ancestor_revision Y',
        #                           {'x': fromeid, 'y': toeid}):
        #     return
        # session.repo.glob_add_relation(session,
        #                                fromeid, 'ancestor_revision', toeid)
        # need NOT safety belt in case of merge nodes with two parents
        # session.unsafe_execute(
        #     'SET R ancestor_revision AR '
        #     'WHERE R eid %(rev)s, PR eid %(prev)s, PR ancestor_revision AR, '
        #     'NOT R ancestor_revision AR',
        #     {'rev': fromeid, 'prev': toeid}, 'rev')
        session.system_sql('INSERT INTO ancestor_revision_relation(eid_from,eid_to) '
                           'SELECT %s, %s '
                           'WHERE NOT EXISTS('
                           '  SELECT 1 FROM ancestor_revision_relation AS ar '
                           '  WHERE ar.eid_from=%s AND ar.eid_to=%s'
                           ')' % (fromeid, toeid, fromeid, toeid))
        session.system_sql('INSERT INTO ancestor_revision_relation(eid_from,eid_to) '
                           'SELECT %s, ar.eid_to FROM ancestor_revision_relation AS ar '
                           'WHERE ar.eid_from=%s AND NOT EXISTS('
                           '  SELECT 1 FROM ancestor_revision_relation AS ar2 '
                           '  WHERE ar2.eid_from=%s AND ar2.eid_to=ar.eid_to'
                           ')' % (fromeid, toeid, fromeid))

# safety belts #################################################################

def _check_in_transaction(vf_or_rev):
    """check that a newly added versioned file or revision entity is done in
    a vcs repository transaction.
    """
    try:
        vcsrepo = vf_or_rev.from_repository[0]
    except IndexError:
        raise missing_relation_error(vf_or_rev, 'from_repository')
    try:
        transactions = vcsrepo.req.transaction_data['vctransactions']
        transaction = transactions[vcsrepo.eid]
    except KeyError:
        raise QueryError('no transaction in progress for repository %s'
                         % vcsrepo.eid)
    return transaction


class CheckVersionedFileOp(PreCommitOperation):
    """check transaction consistency when adding new revision using rql queries
    """
    def precommit_event(self):
        _check_in_transaction(self.entity)


class CheckRevisionOp(PreCommitOperation):
    """check transaction consistency when adding new revision using rql queries
    """
    def precommit_event(self):
        try:
            revision = self.entity.from_revision[0]
        except IndexError:
            raise missing_relation_error(self.entity, 'from_revision')
        transaction = _check_in_transaction(revision)
        if not transaction.reveid == revision.eid:
            raise QueryError('entity linked to a bad revision')


class CheckImmutalbeAttributeHook(Hook):
    events = ('before_update_entity',)
    accepts = ('Revision', 'DeletedVersionContent', 'VersionContent')

    def call(self, session, entity):
        for attr in entity.keys():
            if attr == 'eid':
                continue
            if '%s.%s' % (entity.id, attr) in IMMUTABLE_ATTRIBUTES:
                raise QueryError('%s attribute is not editable' % attr)


class UpdateRepositoryHook(Hook):
    """add repository eid to vcs bridge cache"""
    events = ('before_update_entity', )
    accepts = ('Repository',)
    def call(self, session, entity):
        # XXX check value actually changed
        try:
            edited = entity.edited_attributes
        except AttributeError:
            edited = entity
        if 'path' in edited:
            msg = session._('updating path attribute of a repository isn\'t '
                            'supported. Delete it and add a new one.')
            raise QueryError(msg)
        if 'type' in edited:
            msg = session._('updating type attribute of a repository isn\'t '
                            'supported. Delete it and add a new one.')
            raise QueryError(msg)


# folder/tag extensions ########################################################

class ClassifyVersionedFileHook(Hook):
    """classifies VersionedFile automatically according to their path in the
    repository (require cubicweb-tag and/or cubicweb-folder installed)
    """
    events = ('after_add_entity', )
    accepts = ('VersionedFile',)

    def call(self, session, entity):
        try:
            rschema = self.schema['tags']
            support_tags = rschema.has_rdef('Tag', entity.e_schema)
        except KeyError:
            support_tags = False
        try:
            rschema = self.schema['filed_under']
            support_folders = rschema.has_rdef(entity.e_schema, 'Folder')
        except KeyError:
            support_folders = False
        if not (support_tags or support_folders):
            return
        for directory in entity.directory.split(os.sep):
            if not directory:
                continue
            if support_tags:
                rset = session.execute('Tag X WHERE X name %(name)s',
                                       {'name': directory})
                if rset:
                    session.execute('SET T tags X WHERE X eid %(x)s, T eid %(t)s',
                                    {'x': entity.eid, 't': rset[0][0]}, 'x')
            if support_folders:
                rset = session.execute('Folder X WHERE X name %(name)s',
                                       {'name': directory})
                if rset:
                    session.execute('SET X filed_under F WHERE X eid %(x)s, F eid %(f)s',
                                    {'x': entity.eid, 'f': rset[0][0]}, 'x')



# -*- coding: utf-8 -*-
"""makes the bridge between repository handlers and related cubicweb entities

:organization: Logilab
:copyright: 2007-2010 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""
__docformat__ = "restructuredtext en"

from logilab.common.modutils import LazyObject
from logilab.mtconverter import guess_mimetype_and_encoding
from yams import ValidationError

from cubicweb import QueryError, typed_eid
from cubes.vcsfile import queries

_VCSTYPES = {'subversion': LazyObject('cubes.vcsfile.reposvn', 'SVNRepository'),
             'mercurial': LazyObject('cubes.vcsfile.repohg', 'HGRepository'),
             }
_REPOHDLRS = {}


class VCSException(Exception):
    def __init__(self, repoeid, attr, msgid, msgargs):
        self.repoeid = repoeid
        self.attr = attr
        self.msgid = msgid
        self.msgargs = msgargs

    def __str__(self):
        return '%s.%s: %s' % (self.repoeid, self.attr,
                              self.msgid % self.msgargs)

    def to_validation_error(self, translate):
        msg = translate(self.msgid) % self.msgargs
        return ValidationError(self.repoeid, {self.attr: msg})


def repository_handler(repoeid):
    """get repository handler for the repository with the given eid"""
    return _REPOHDLRS[repoeid]


def cache_handler_for_repository(repoent):
    """register a repository handler"""
    try:
        repohdlr = _REPOHDLRS[repoent.eid]
        repohdlr.encoding = repoent.encoding # in case it changed
        return repohdlr
    except KeyError:
        pass
    repohdlr = build_repository_handler(repoent.eid, repoent.type, repoent.path,
                                        repoent.encoding, repoent.req._)
    _REPOHDLRS[repoent.eid] = repohdlr
    return repohdlr


def build_repository_handler(eid, type, path, encoding, _=unicode):
    """instantiate a repository handler"""
    try:
        repohdlr = _VCSTYPES[type](eid, path, encoding)
    except KeyError:
        msg = _('%s is not a known repository type')
        raise ValidationError(eid, {'type': msg % type})
    except ImportError:
        msg = _('missing python bindings to support %s repositories')
        raise ValidationError(eid, {'type': msg % type})
    _REPOHDLRS[eid] = repohdlr
    return repohdlr


def import_content(repo, commitevery=10):
    """synchronize content of known vcs repositories

    `repo` is the cubicweb repository (not a vcs repository)
    """
    session = repo.internal_session()
    try:
        for repoentity in session.execute(
            'Any X,T,P,SP,E WHERE X is Repository, '
            'X type T, X path P, X subpath SP, X encoding E').entities():
            # give a change to threads waiting for a pool
            session.set_pool()
            try:
                repohdlr = cache_handler_for_repository(repoentity)
            except VCSException, ex:
                repo.error(str(ex))
                continue
            try:
                repohdlr.import_content(repoentity, commitevery)
                session.commit()
            except:
                import traceback
                traceback.print_exc()
                repoentity.exception(
                    'error while importing content for vcs repo %s',
                    repoentity.dc_title())
                session.rollback()
    finally:
        session.close()


def set_at_revision(session, reid, safetybelt=False):
    """at_revision is an internal relation used to know the full content of a
    repository at any revision. This function set this relation for newly
    created revision.
    """
    rqlexec = session.unsafe_execute
    # link to version content generated by this revision
    #rql = 'SET VC at_revision REV WHERE REV eid %(rev)s, VC from_revision REV'
    #if safetybelt:
    #    rql += ', NOT VC at_revision REV'
    #rqlexec(rql, {'rev': reid}, 'rev')
    sqlrestr = ('SELECT VC.cw_eid, %s FROM cw_%%s AS VC '
                'WHERE VC.cw_from_revision=%s' % (reid, reid))
    if safetybelt:
        sqlrestr += ('AND NOT EXISTS(SELECT 1 FROM at_revision_relation AS atr '
                     'WHERE atr.eid_to=%s AND atr.eid_from=VC.cw_eid)') % reid

    sql = ('INSERT INTO at_revision_relation(eid_from,eid_to) '
           '%s UNION %s') % (sqlrestr % 'VersionContent',
                             sqlrestr % 'DeletedVersionContent')
    session.system_sql(sql)
    # link to version content of files in parent revisions but not modified by
    # this revision
    # rql = ('SET VC at_revision REV WHERE VC is VersionContent, '
    #        'REV eid %(rev)s, VC at_revision PREV, '
    #        'REV parent_revision PREV, '
    #        'NOT EXISTS(VC2 from_revision REV, VC content_for VF, '
    #        'VC2 content_for VF)')
    # if safetybelt:
    #     rql += ', NOT VC at_revision REV'
    # rqlexec(rql, {'rev': reid}, 'rev')
    sql = ('INSERT INTO at_revision_relation(eid_from,eid_to) '
           'SELECT DISTINCT VC.cw_eid, %s '
           'FROM at_revision_relation AS ar, parent_revision_relation AS pr, cw_VersionContent AS VC '
           'WHERE VC.cw_eid=ar.eid_from AND pr.eid_from=%s AND pr.eid_to=ar.eid_to '
           'AND NOT EXISTS('
           '  SELECT 1 FROM cw_VersionContent AS VC2 '
           '  WHERE VC2.cw_from_revision=%s AND VC2.cw_content_for=VC.cw_content_for '
           '  UNION '
           '  SELECT 1 FROM cw_DeletedVersionContent AS VC2 '
           '  WHERE VC2.cw_from_revision=%s AND VC2.cw_content_for=VC.cw_content_for)') % (
        reid, reid, reid, reid)
    if safetybelt:
        sql += (' AND NOT EXISTS(SELECT 1 FROM at_revision_relation AS atr '
                'WHERE atr.eid_from=%s AND atr.eid_to=VC.cw.cw_eid)') % reid
    session.system_sql(sql)


def import_revision(session, repoeid, date, **kwargs):
    """import a new revision from a repository"""
    args = {'date': date}
    args['repoeid'] = repoeid
    for attr in ('revision', 'author', 'description', 'changeset', 'branch'):
        args[attr] = kwargs.get(attr)
    prevs = kwargs['parents']
    if prevs:
        args['parent'] = prevs[0]
    reveid = session.execute(queries.new_revision_rql(prevs, True), args)[0][0]
    if len(prevs) > 1:
        for preveid in prevs[1:]:
            session.execute('SET R parent_revision PR WHERE R eid %(r)s, PR eid %(pr)s',
                            {'r': reveid, 'pr': preveid})
    return reveid

def import_versioned_file(session, repoeid, date, directory, name):
    """import a version controlled file from a repository"""
    args = {'directory': directory, 'name': name, 'repoeid': repoeid}
    return session.execute(
        'INSERT VersionedFile X: X directory %(directory)s, '
        'X name %(name)s, X from_repository R '
        'WHERE R eid %(repoeid)s', args)[0][0]

def import_version_content(session, repoeid, reveid, upath, date, **kwargs):
    """import a new file revision from a repository"""
    repohdlr = repository_handler(repoeid)
    vfeid = repohdlr.vf_eid(session, upath, date)
    # done in a hook but doing it here as well to avoid an additional query to
    # lookup filename
    mt, enc = guess_mimetype_and_encoding(
        fallbackencoding=repohdlr.encoding, fallbackmimetype=None,
        filename=upath)
    args = {'vf': vfeid, 'rev': reveid,
            'df': mt and unicode(mt) or None,
            'de': enc and unicode(enc) or None}
    rql = ('INSERT VersionContent X: X from_revision REV, X content_for VF, '
           'X data_format %(df)s, X data_encoding %(de)s')
    for inlinerel in ('vc_copy', 'vc_rename'):
        if inlinerel in kwargs:
            args[inlinerel] = kwargs[inlinerel]
            rql += ', X %s %%(%s)s' % (inlinerel, inlinerel)
    rql += ' WHERE REV eid %(rev)s, VF eid %(vf)s'
    return session.execute(rql, args)[0][0]


def import_deleted_version_content(session, repoeid, reveid, upath, date):
    """import a new deleted file revision from a repository"""
    repohdlr = repository_handler(repoeid)
    vfeid = repohdlr.vf_eid(session, upath, date)
    args = {'vf': vfeid, 'rev': reveid}
    rql = ('INSERT DeletedVersionContent X: X from_revision REV, X content_for VF '
           ' WHERE REV eid %(rev)s, VF eid %(vf)s')
    return session.execute(rql, args)[0][0]



"""repository abstraction for supported repositories

:organization: Logilab
:copyright: 2007-2009 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""
__docformat__ = "restructuredtext en"
_ = unicode

from os.path import exists, split, join
from logging import getLogger

from cubicweb import set_log_methods

from cubes.vcsfile import bridge

class VCSRepository(object):
    type = None # to set in concret classes

    def __init__(self, eid, path, encoding):
        # eid of the associate Repository entity
        self.eid = eid
        # encoding of files, messages... Normalize for later comparison
        self.encoding = encoding.lower()
        # path to the repository
        if isinstance(path, unicode):
            fspath = path.encode(encoding)
        else:
            fspath = path
        if not exists(fspath):
            msg = _('%s does not exist on the file system')
            raise bridge.VCSException(eid, 'path', msg, path)
        try:
            fspath = self.canonicalize(fspath)
        except:
            self.exception('invalid repository')
            msg = _('%(repopath)s is not a valid %(repotype)s repository')
            raise bridge.VCSException(eid, 'path', msg, {'repopath': path,
                                                         'repotype': self.type})
        self.path = fspath
        self.vf_path_cache = {}

    def __str__(self):
        return '%s:%s' % (self.type, self.path)

    def encode(self, ustring):
        if ustring:
            return ustring.encode(self.encoding)
        return ''

    def decode(self, string):
        if string:
            return string.decode(self.encoding)
        return u''

    def canonicalize(self, path):
        return path

    def import_content(self, repoentity):
        """import content from the repository"""
        raise NotImplementedError()

    def vf_eid(self, session, path, date):
        """get versioned file eid for the given path, inserting it if necessary
        """
        try:
            return self.vf_path_cache[path]
        except KeyError:
            directory, fname = split(path)
            rset = session.unsafe_execute(
                'Any X WHERE X directory %(dir)s, X name %(name)s, '
                'X from_repository R, R eid %(r)s',
                {'dir': directory, 'name': fname, 'r': self.eid})
            if rset:
                eid = rset[0][0]
            else:
                eid = bridge.import_versioned_file(session, self.eid, date,
                                                   directory, fname)
            self.vf_path_cache[path] = eid
            return eid

    def file_content(self, directory, fname, rev):
        """extract a binary string with the content of the file at `directory` /
        `fname` for revision `rev` in the repository
        """
        if isinstance(fname, unicode):
            fname = self.encode(fname)
        if directory is not None:
            if isinstance(directory, unicode):
                directory = self.encode(directory)
            path = join(directory, fname)
        else:
            path = fname
        return self._file_content(path, rev)

    def _file_content(self, path, rev):
        raise NotImplementedError()

    def revision_transaction(self, session, entity):
        """open a transaction to create a new revision corresponding to the
        given entity
        """
        raise NotImplementedError()

    def add_versioned_file_content(self, session, transaction, vf, entity,
                                   data):
        """add a new revision of a versioned file"""
        raise NotImplementedError()

    def add_versioned_file_deleted_content(self, session, transaction, vf,
                                           entity):
        """add a new revision of a just deleted versioned file"""
        raise NotImplementedError()

set_log_methods(VCSRepository, getLogger('cubicweb.vcs'))



from time import time

add_relation_type('ancestor_revision')

rset = rql('Revision X ORDERBY R WHERE X revision R', ask_confirm=False)
for reid, in rset:
    print 'set ancestor for', reid,
    t = time()
    rql('SET REV ancestor_revision PREV WHERE REV eid %(rev)s, REV parent_revision PREV',
        {'rev': reid}, 'rev', ask_confirm=False)
    rql('SET REV ancestor_revision AREV WHERE REV eid %(rev)s, REV parent_revision PREV, '
        'PREV ancestor_revision AREV', {'rev': reid}, 'rev', ask_confirm=False)
    print '%.3f sec.' % (time() - t)

checkpoint()



# pylint: disable-msg=W0622
"""cubicweb-vcsfile packaging information"""

modname = 'vcsfile'
distname = "cubicweb-%s" % modname

numversion = (1, 0, 13)
version = '.'.join(str(num) for num in numversion)

copyright = '''Copyright (c) 2007-2010 LOGILAB S.A. (Paris, FRANCE).
http://www.logilab.fr/ -- mailto:contact@logilab.fr'''

license = 'LGPL'
author = "Logilab"
author_email = "contact@logilab.fr"
web = 'http://www.cubicweb.org/project/%s' % distname

short_desc = "component to use version control systems from the CubicWeb framework"
long_desc = """CubicWeb VCSFile component to integrate with version control systems and
make their data available from within a CubicWeb application (currently
works for mercurial and subversion).

CubicWeb is a semantic web application framework, see http://www.cubicweb.org
"""

from os import listdir as _listdir
from os.path import join, isdir
from glob import glob

THIS_CUBE_DIR = join('share', 'cubicweb', 'cubes', modname)

def listdir(dirpath):
    return [join(dirpath, fname) for fname in _listdir(dirpath)
            if fname[0] != '.' and not fname.endswith('.pyc')
            and not fname.endswith('~')
            and not isdir(join(dirpath, fname))]

data_files = [
    # common files
    [THIS_CUBE_DIR, [fname for fname in glob('*.py') if fname != 'setup.py']],
    ]
# check for possible extended cube layout
for dirname in ('entities', 'views', 'sobjects', 'hooks', 'schema', 'data', 'i18n', 'migration'):
    if isdir(dirname):
        data_files.append([join(THIS_CUBE_DIR, dirname), listdir(dirname)])
# Note: here, you'll need to add subdirectories if you want
# them to be included in the debian package

cube_eid = None # XXX
# used packages
__depends_cubes__ = {}
__depends__ = {'cubicweb': '>= 3.3.1',
               'logilab-mtconverter': '>= 0.7.0',
               'pysqlite': None,
               }
__use__ = tuple(__depends_cubes__)
classifiers = [
           'Environment :: Web Environment',
           'Framework :: CubicWeb',
           'Programming Language :: Python',
           'Programming Language :: JavaScript',
           ]



"template automatic tests"
import os
from logilab.common.testlib import unittest_main
from cubicweb.devtools import BaseApptestConfiguration, testlib

class MyConfig(BaseApptestConfiguration):
    sourcefile = 'sources'

class AutomaticWebTest(testlib.AutomaticWebTest):
    configcls = MyConfig
    no_auto_populate = ('Repository', 'Revision', 'VersionedFile',
                        'VersionContent', 'DeletedVersionContent',)
    ignored_relations = ('at_revision', 'parent_revision',
                         'from_repository', 'from_revision', 'content_for',)

    def custom_populate(self, how_many, cursor):
        self.repo = self.add_entity('Repository', type=u'mercurial',
                                    path=u'testrepohg', encoding=u'latin1')

if __name__ == '__main__':
    unittest_main()
